/*
 * The Pipelined Regular Expression Monitor Compiler System
 * Copyright (c) 2002 University of British Columbia
 * All Rights Reserved.
 *
 *
 * PREMSC - A production based monitor compiler.  
 *
 * Marcio Oliveira -- 2002
 */

#include <stdio.h>
#include <stdlib.h>
#include <error.h>
#include <compile.h>
#include <translation.h>
#include <verilog.h>
#include <vhdl.h>

#define VERILOG 1
#define VHDL    2

extern int strcasecmp();

extern int srcline;
extern char *yytext;
extern FILE *yyin;
extern int yydebug;



int verbose = 0;
int language = VERILOG;
char *inFile = NULL;
char *outFile = NULL; 

void yyerror(char *s) {
    printf("\n%s at line %d, token %s.\n\n", s, srcline, yytext);
    exit(1);
};

int yywrap() {
    return(1);
};

void yyparse();

extern FILE *yyin;
extern int yydebug;

FILE *openOutputFile(char *outFile) {
    
    FILE *fd;

    if (outFile == NULL) {
        fd = stdout;
    } else {
        fd = fopen(outFile, "w");
        if (fd == NULL)
            fd = stdout;
    }

    return fd;
}

void print_usage(char *name) {
    printf("\n\n*** PREMS ***\n\n");
    printf("Usage: %s [options] <filename>, where options are:\n", name);
    printf("   -t <verilog, vhdl>: Set output language [default verilog];\n");
    printf("   -o <file>:          Write output to <file> [default stdout];\n");
    printf("\n\n");
}

int main (int argc, char *argv[]) {

    int i;
    FILE *out; 

    if (argv[1] == 0) {
        print_usage(argv[0]);
        exit(1);
    };

    i = 1;
    while (argv[i][0] == '-') {
        switch (argv[i][1]) {
            case 'v':
                verbose = 1;
                break;
            case 't':
                i++;
                if (!strcasecmp("verilog", argv[i])) {
                    language = VERILOG;
                } else if (!strcasecmp("vhdl", argv[i])) {
                    language = VHDL;
                } else {
                    print_usage(argv[0]);
                    exit(1);
                }  
                break;
            case 'o':
                i++;
                outFile = argv[i];
                break;
            default: 
                print_usage(argv[0]);
                exit(1); 
        }
        i++;
    }

    inFile = argv[i];
   
    if ((yyin = fopen(inFile, "r")) == NULL) {
        fatal_error("Invalid file name %s.\n", inFile);
    }

    initCompile();
    yydebug = 0;
    yyparse();

    out = openOutputFile(outFile);

    switch (language) {
        case VHDL:
            dumpVhdl(out); 
            break;
        case VERILOG:
            dumpVerilog(out);
            break;
    }  

    fclose(out);

    return 1;
};
