/*
 * The Production Based Monitor Compiler System
 * Copyright (c) 2002 University of British Columbia
 * All Rights Reserved.
 *
 *
 * Lexical Analyzer for PBC
 *
 * Marcio Oliveira -- 2002
 */

%{

#include <symbols.h>
#include <compile.h>
#include <y.tab.h>

int srcline = 1;

%}

%%

[ \t]              ;
\n                 srcline++;

"/*"               {
                     char c1, c2;
                     /*
                      * When a comment start is encountered, skip
                      * everything until the close of the comment
                      */
                     c1 = c2 = input();
                     if (c1 == '\n') srcline++;
                     while (!(c1 == '*' && c2 == '/')){
                       c1 = c2;
                       c2 = input();
                       if (c2 == '\n') srcline++;
                     }
                   };

"("        return(LP);
")"        return(RP);

"{"        return(LCB);
"}"        return(RCB);

"["        return(LB);
"]"        return(RB); 
":"        return(COLON);
".."       return(DOTDOT);  

"="        return(EQUAL);

","        return(COMMA);
"||"       return(ORP);
"^"        return(HAT);
"*"        return(STAR);
"+"        return(PLUS);
"@"        return(AT);

"-"        return(MINUS);

";"        return(SEMICOLON);

"!"        return(NOTB);
"&"        return(ANDB);
"|"        return(ORB);

"->"       return(RARROW);
"<-"       return(LARROW);
"=="       return(EQUALEQUAL);
"!="       return(NOTEQUAL);

"internal" return(INTERNAL);
"input"    return(INPUT);
"output"   return(OUTPUT);
"in_out"   return(IN_OUT);

"define"   return(DEFINE);  

"monitor"  return(MONITOR); 

[_A-Za-z][A-Za-z0-9._\$-]* {
                     yylval.string_t = (char *) malloc(strlen(yytext) + 1);
                     strcpy(yylval.string_t, yytext);
                     return(IDENTIFIER);
                   };
[0-9]+             {
                     sscanf(yytext,"%d",&yylval.int_t);
                     return(CONSTANT);
                   };

. {
    fprintf(stderr, "Syntax error at line: %d. Symbol: %s\n", srcline, yytext);
    exit(1);
};

%%
