/*                     
 * The Pipelined Regular Expression Monitor Compiler System
 * Copyright (c) 2002 University of British Columbia
 * All Rights Reserved.
 *                 
 *                     
 * VHDL code generation.   
 *
 * Marcio Oliveira -- 2002
 */

#include <vhdl.h>
#include <translation.h>
#include <symbols.h>
#include <error.h>

extern stateMachinePtr stateMachineList;

extern signalPtr signalList;
extern signalPtr internalList;
extern definePtr defineList;

extern controlPointPtr controlPointList;
extern int controlPointCount;

extern actionPointPtr actionPointList;
extern int actionPointCount;

extern exThreadPtr exThreadList;
extern int exThreadCount;

extern wirePtr wireList;
extern int wireCount;

void dumpVhdl(FILE *fd) {
    
    dumpVhdlEntity(fd);
    dumpVhdlArchitecture(fd);
} 

void dumpVhdlEntity(FILE *fd) {

    signalPtr s;

    fprintf(fd, "entity MONITOR is\n");
    fprintf(fd, "    port(");

    if (signalList != NULL) {
        s = signalList;
        if (s->isArray) {
            if (s->index[0] < s->index[1]) {
                fprintf(fd, "%s : in BIT_VECTOR(%d to %d);\n",
                        s->name, s->index[0], s->index[1]);
            } else {
                fprintf(fd, "%s : in BIT_VECTOR(%d downto %d);\n",
                        s->name, s->index[0], s->index[1]);
            } 
        } else {
            fprintf(fd, "%s : in BIT;\n", s->name);
        }
    }

    for (s = signalList->next; s != NULL; s = s->next) {
        if (s->isArray) {
            if (s->index[0] < s->index[1]) {
                fprintf(fd, "         %s : in BIT_VECTOR(%d to %d);\n", 
                        s->name, s->index[0], s->index[1]);
            } else {
                fprintf(fd, "         %s : in BIT_VECTOR(%d downto %d);\n", 
                        s->name, s->index[0], s->index[1]);
            } 
        } else {
            fprintf(fd, "         %s : in BIT;\n", s->name);
        }
    }

    fprintf(fd, "         clk, reset: in BIT;\n");
    fprintf(fd, "         ok : out BIT);\n");

    fprintf(fd, "end MONITOR;\n\n");
}

void dumpVhdlArchitecture(FILE *fd) {

    wirePtr w;
    definePtr d;
    signalPtr s;

    fprintf(fd, "architecture MONITOR_BEHAVIOUR of MONITOR is\n");
    fprintf(fd, "    signal pbs_w : BIT_VECTOR(0 to %d);\n", wireCount-1);
    fprintf(fd, "    signal pbs_X : BIT_VECTOR(%d downto 0);\n",
            controlPointCount-1);
    for (d = defineList; d != NULL; d = d->next) {
        fprintf(fd, "    signal %s : BIT;\n", d->name);
    }

    for (s = internalList; s != NULL; s = s->next) {
        if (s->isArray) {
            if (s->index[0] < s->index[1]) {
                fprintf(fd, "    signal %s : BIT_VECTOR(%d to %d);\n",
                        s->name, s->index[0], s->index[1]);
            } else {
                fprintf(fd, "    signal %s : BIT_VECTOR(%d downto %d);\n", s->name, s->index[0], s->index[1]);
            }
        } else {
            fprintf(fd, "    signal %s : BIT;\n", s->name);
        }
    }

    fprintf(fd, "\n");

    fprintf(fd, "    function BITS (NAT: NATURAL; BITS: BIT_VECTOR) return BIT_VECTOR is\n");
    fprintf(fd, "        variable TEMP : NATURAL;\n");
    fprintf(fd, "        variable RESULT : BIT_VECTOR(BITS'RANGE);\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        TEMP := NAT;\n");
    fprintf(fd, "        if (BITS'ASCENDING) then\n");
    fprintf(fd, "            for INDEX in BITS'REVERSE_RANGE loop\n");
    fprintf(fd, "                RESULT(INDEX) := bit'val(TEMP rem 2);\n");
    fprintf(fd, "                TEMP := TEMP / 2;\n");
    fprintf(fd, "            end loop;\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            for INDEX in BITS'RANGE loop\n");
    fprintf(fd, "                RESULT(INDEX) := bit'val(TEMP rem 2);\n");
    fprintf(fd, "                TEMP := TEMP / 2;\n");
    fprintf(fd, "            end loop;\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end BITS;\n\n");

    fprintf(fd, "    function BITS (NAT: NATURAL; BITS: BIT) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (NAT = 1) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end BITS;\n\n");

    fprintf(fd, "    function INT(T: BIT_VECTOR) return NATURAL is\n");
    fprintf(fd, "        variable RESULT: NATURAL;\n");
    fprintf(fd, "        variable COUNT: NATURAL;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        RESULT := 0;\n");
    fprintf(fd, "        COUNT := 0;\n");
    fprintf(fd, "        if (T'ASCENDING) then\n");
    fprintf(fd, "            for INDEX in T'RANGE loop\n");
    fprintf(fd, "                if (T(INDEX) = '1') then\n");
    fprintf(fd, "                    RESULT := RESULT + (2**COUNT);\n");
    fprintf(fd, "                end if;\n");
    fprintf(fd, "                COUNT := COUNT + 1;\n");
    fprintf(fd, "            end loop;\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            for INDEX in T'REVERSE_RANGE loop\n");
    fprintf(fd, "                if (T(INDEX) = '1') then\n");
    fprintf(fd, "                    RESULT := RESULT + (2**COUNT);\n");
    fprintf(fd, "                end if;\n");
    fprintf(fd, "                COUNT := COUNT + 1;\n");
    fprintf(fd, "            end loop;\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end INT;\n\n");

    fprintf(fd, "    function INT(T: BIT) return NATURAL is\n");
    fprintf(fd, "        variable RESULT: NATURAL;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (T = '1') then\n");
    fprintf(fd, "            RESULT := 1;\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := 0;\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end INT;\n\n");

    fprintf(fd, "    function \"=\"(OP1, OP2: BIT_VECTOR) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (OP1 = OP2) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"=\";\n\n");

    fprintf(fd, "    function \"=\"(OP1, OP2: BIT) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (OP1 = OP2) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"=\";\n\n");

    fprintf(fd, "   function \"=\"(OP1: BIT; OP2: NATURAL) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (INT(OP1) = OP2) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"=\";\n\n");

    fprintf(fd, "    function \"=\"(OP1: NATURAL; OP2: BIT) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (OP1 = INT(OP2)) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"=\";\n\n");

    fprintf(fd, "    function \"=\"(OP1: BIT_VECTOR; OP2: NATURAL) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (INT(OP1) = OP2) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"=\";\n\n");

    fprintf(fd, "    function \"=\"(OP1: NATURAL; OP2: BIT_VECTOR) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (OP1 = INT(OP2)) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"=\";\n\n");

    fprintf(fd, "    function \"/=\"(OP1, OP2: BIT_VECTOR) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (OP1 /= OP2) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"/=\";\n\n");

    fprintf(fd, "    function \"/=\"(OP1, OP2: BIT) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (OP1 /= OP2) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"/=\";\n\n");

    fprintf(fd, "   function \"/=\"(OP1: BIT; OP2: NATURAL) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (INT(OP1) /= OP2) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"/=\";\n\n");

    fprintf(fd, "    function \"/=\"(OP1: NATURAL; OP2: BIT) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (OP1 /= INT(OP2)) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"/=\";\n\n");

    fprintf(fd, "    function \"/=\"(OP1: BIT_VECTOR; OP2: NATURAL) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (INT(OP1) /= OP2) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"/=\";\n\n");

    fprintf(fd, "    function \"/=\"(OP1: NATURAL; OP2: BIT_VECTOR) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (OP1 /= INT(OP2)) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"/=\";\n\n");

    fprintf(fd, "    function \"/=\"(OP1, OP2: NATURAL) return BIT is\n");
    fprintf(fd, "        variable RESULT: BIT;\n");
    fprintf(fd, "    begin\n");
    fprintf(fd, "        if (OP1 /= OP2) then\n");
    fprintf(fd, "            RESULT := '1';\n");
    fprintf(fd, "        else\n");
    fprintf(fd, "            RESULT := '0';\n");
    fprintf(fd, "        end if;\n");
    fprintf(fd, "\n");
    fprintf(fd, "        return RESULT;\n");
    fprintf(fd, "    end \"/=\";\n\n");

    fprintf(fd, "begin\n\n");

    for (d = defineList; d != NULL; d = d->next) {
        fprintf(fd, "    %s <= ", d->name);
        dumpVhdlExpression(fd, d->expr, "X");
        fprintf(fd, ";\n");
    }
    if (defineList != NULL)
        fprintf(fd, "\n");

    for (w = wireList; w != NULL; w = w->next) {
        fprintf(fd, "    pbs_w(%d) <= ", w->index);
        dumpVhdlExpression(fd, w->f, "X");
        fprintf(fd, ";\n");
    }
    fprintf(fd, "\n");

    dumpVhdlStateMachine(fd);

    fprintf(fd, "end MONITOR_BEHAVIOUR;\n\n");
}

void dumpVhdlStateMachine(FILE *fd) {

    fprintf(fd, "    STATE_MACHINE: process(clk)\n");

    dumpVhdlStateMachineVars(fd);

    fprintf(fd, "    begin\n\n");
    fprintf(fd, "        if (clk = '1') and (reset = '1') then\n");

    dumpVhdlStateMachineInit(fd);

    fprintf(fd, "        elsif (clk = '1') then\n");

    dumpVhdlStateMachineStep(fd);

    fprintf(fd, "        end if;\n\n");
    fprintf(fd, "    end process STATE_MACHINE;\n\n");
}

void dumpVhdlStateMachineVars(FILE *fd) {

    fprintf(fd, "        variable pbs_oki : BIT;\n");
    fprintf(fd, "        variable pbs_active : BIT_VECTOR(%d downto 0);\n", 
            exThreadCount-1);
    fprintf(fd, "        variable pbs_enable : BIT_VECTOR(%d downto 0);\n", 
            exThreadCount-1);
    fprintf(fd, "        variable pbs_okn : BIT_VECTOR(%d downto 0);\n", 
            exThreadCount-1);
    fprintf(fd, "        variable pbs_reentry : BIT_VECTOR(%d downto 0);\n", 
            exThreadCount-1);
    fprintf(fd, "        variable pbs_ok : BIT_VECTOR(%d downto 0);\n", 
            exThreadCount-1);
    fprintf(fd, "        variable pbs_Y : BIT_VECTOR(%d downto 0);\n", 
            controlPointCount-1);

    if (actionPointCount != 0)
        fprintf(fd, "        variable pbs_A : BIT_VECTOR(%d downto 0);\n", 
                actionPointCount-1);

}

void dumpVhdlStateMachineInit(FILE *fd) {

    controlPointPtr cp;
    signalPtr s;

    for (cp = controlPointList; cp != NULL; cp = cp->next) {
        fprintf(fd, "            pbs_X(%d) <= ", cp->index);
        if (cp->f == NULL)
            fprintf(fd, "'1';\n");
        else
            fprintf(fd, "'0';\n");
    }

    fprintf(fd, "            pbs_oki := '1';\n");
    fprintf(fd, "            ok <= pbs_oki;\n");

    for (s = internalList; s != NULL; s = s->next) {
        if (s->isArray) {
            fprintf(fd, "            %s <= BITS(%d, %s);\n", s->name,
                                                             s->initValue,
                                                             s->name); 
        } else {
            fprintf(fd, "            %s <= '%d';\n", s->name, s->initValue);
        }   
    }
}

void dumpVhdlStateMachineStep(FILE *fd) {

    controlPointPtr cp;
    exThreadPtr et;
    actionPointPtr ap;
    stateMachinePtr sm;

    for (cp = controlPointList; cp != NULL; cp = cp->next) {
        fprintf(fd, "            pbs_Y(%d) := ", cp->index);
        dumpVhdlExpression(fd, cp->f, "X");
        fprintf(fd, ";\n");
    }
    fprintf(fd, "\n");

    for (et = exThreadList; et != NULL; et = et->next) {
        fprintf(fd, "            pbs_active(%d) := ", et->index);
        dumpVhdlExpression(fd, et->active, "X");
        fprintf(fd, ";\n");
    }
    fprintf(fd, "\n");

    for (et = exThreadList; et != NULL; et = et->next) {
        fprintf(fd, "            pbs_okn(%d) := ", et->index);
        dumpVhdlExpression(fd, et->rec, "X");
        fprintf(fd, ";\n");
    }
    fprintf(fd, "\n"); 

    for (et = exThreadList; et != NULL; et = et->next) {
        fprintf(fd, "            pbs_reentry(%d) := pbs_active(%d) and pbs_okn(%d);\n", et->index, et->index, et->index);
    }
    fprintf(fd, "\n");

    for (et = exThreadList; et != NULL; et = et->next) {
        fprintf(fd, "            pbs_enable(%d) := ", et->index);
        dumpVhdlExpression(fd, et->enable, "X");
        fprintf(fd, ";\n");
    }
    fprintf(fd, "\n");

    for (et = exThreadList; et != NULL; et = et->next) {
        fprintf(fd, "            pbs_ok(%d) := ", et->index);
        dumpVhdlExpression(fd, et->ok, "X");
        fprintf(fd, ";\n");
    }
    fprintf(fd, "\n");

    for (ap = actionPointList; ap != NULL; ap = ap->next) {
        fprintf(fd, "            pbs_A(%d) := ", ap->index);
        dumpVhdlExpression(fd, ap->f, "X");
        fprintf(fd, ";\n");
    }
    if (actionPointList != NULL)
        fprintf(fd, "\n");

    for (ap = actionPointList; ap != NULL; ap = ap->next) {
        fprintf(fd, "            if (pbs_A(%d) = '1') then\n", ap->index);
        dumpVhdlActionList(fd, ap->actionList);
        fprintf(fd, "            end if;\n\n");
    }
    if (actionPointList != NULL)
        fprintf(fd, "\n");


    for (et = exThreadList; et != NULL; et = et->next) {
        fprintf(fd, "            pbs_oki := pbs_oki and (pbs_ok(%d) or (not pbs_enable(%d))) and (not pbs_reentry(%d));\n", et->index, et->index, et->index);
    }

    for (sm = stateMachineList; sm != NULL; sm = sm->next) {
        fprintf(fd, "            pbs_oki := pbs_oki and (");
        fprintf(fd, "INT(pbs_Y(%d downto %d)) /= 0", sm->cp[1], sm->cp[0]); 
        fprintf(fd, ");\n");
    }

    fprintf(fd, "\n");
    fprintf(fd, "            ok <= pbs_oki;\n");
    fprintf(fd, "            pbs_X <= pbs_Y;\n\n");
}

void dumpVhdlExpression(FILE *fd, treeNodePtr node, char *type) {

    controlPointPtr cp;
    wirePtr w;

    if (node == NULL) {
        fprintf(fd, "'0'");
        return;
    }

    if (isBoolExpr(node)) {
        dumpVhdlBoolExpression(fd, node);
        return;
    }

    switch (node->type) {
        case P_TN_CP:
            cp = (controlPointPtr) node->data;
            fprintf(fd, "pbs_%s(%d)", type, cp->index);
            break;
        case P_TN_AND_C:
            fprintf(fd, "(");
            dumpVhdlExpression(fd, node->left, type);
            fprintf(fd, " and ");
            dumpVhdlExpression(fd, node->right, type);
            fprintf(fd, ")");
            break;
        case P_TN_OR_C:
            fprintf(fd, "(");
            dumpVhdlExpression(fd, node->left, type);
            fprintf(fd, " or ");
            dumpVhdlExpression(fd, node->right, type);
            fprintf(fd, ")");
            break;
        case P_TN_NOT_C:
            fprintf(fd, "(not ");
            dumpVhdlExpression(fd, node->left, type);
            fprintf(fd, ")");
            break;
        case P_TN_TEMP:
            dumpVhdlExpression(fd, node->left, type);
            break;
        case P_TN_TRUE_C:
            fprintf(fd, "'1'");
            break;
        case P_TN_FALSE_C:
            fprintf(fd, "'0'");
            break;
        case P_TN_WIRE_C:
            w = (wirePtr) node->data;
            fprintf(fd, "pbs_w(%d)", w->index);
            break;
        default:
            fatal_error("dumpVhdlExpression()");
    }
}

void dumpVhdlBoolExpression(FILE *fd, treeNodePtr node) {

    definePtr d;
    signalPtr s;

    switch (node->type) {
        case P_TN_DEFINE:
            d = (definePtr) node->data;
            fprintf(fd, "%s", d->name);
            break;
        case P_TN_SIGNAL:
        case P_TN_INTERNAL:
            s = (signalPtr) node->data;
            fprintf(fd, "%s", s->name);
            break;
        case P_TN_SIGNAL_ARRAY:
        case P_TN_INTERNAL_ARRAY:
            s = (signalPtr) node->data;
            fprintf(fd, "%s(", s->name);
            dumpVhdlIndex(fd, s->indexN);
            fprintf(fd, ")");
            break;
        case P_TN_EQUALEQUAL:
            fprintf(fd, "(");
            dumpVhdlBoolExpression(fd, node->left);
            fprintf(fd, " = ");
            dumpVhdlBoolExpression(fd, node->right); 
            fprintf(fd, ")\n");
            break;
        case P_TN_NOTEQUAL:
            fprintf(fd, "(");
            dumpVhdlBoolExpression(fd, node->left);
            fprintf(fd, " /= ");
            dumpVhdlBoolExpression(fd, node->right);
            fprintf(fd, ")\n");
            break;
        case P_TN_AND_B:
            fprintf(fd, "(");
            dumpVhdlBoolExpression(fd, node->left);
            fprintf(fd, " and ");
            dumpVhdlBoolExpression(fd, node->right);
            fprintf(fd, ")");
            break;
        case P_TN_OR_B:
            fprintf(fd, "(");
            dumpVhdlBoolExpression(fd, node->left);
            fprintf(fd, " or ");
            dumpVhdlBoolExpression(fd, node->right);
            fprintf(fd, ")");
            break;
        case P_TN_NOT_B:
            fprintf(fd, "not ");
            dumpVhdlBoolExpression(fd, node->left);
            break;
        case P_TN_CONSTANT:
            fprintf(fd, "%d", (int) node->data);
            break;
        default:
            fatal_error("dumpVhdlBoolExpression()");
    }
}

void dumpVhdlActionList(FILE *fd, actionPtr a) {

    actionPtr ap;

    for (ap = a; ap != NULL; ap = ap->next) {
        fprintf(fd, "                ");
        dumpVhdlActionExpressionLhs(fd, ap->lhs);
        fprintf(fd, " <= BITS(");
        dumpVhdlActionExpression(fd, ap->rhs);
        fprintf(fd, ", ");
        dumpVhdlActionExpressionLhs(fd, ap->lhs);
        fprintf(fd, ");\n");
    }
}


void dumpVhdlActionExpressionLhs(FILE *fd, actionNodePtr an) {

    switch (an->type) {
        case P_AN_SIGNAL:
            fprintf(fd, "%s", an->data.signal->name);
            break;
        case P_AN_SIGNAL_ARRAY:
            fprintf(fd, "%s(", an->data.signal->name);
            dumpVhdlIndex(fd, an->index);
            fprintf(fd, ")");
            break;
        default:
            fatal_error("dumpVhdlActionExpression()");
    }
}

void dumpVhdlActionExpression(FILE *fd, actionNodePtr an) {

    switch (an->type) {
        case P_AN_PLUS:
            fprintf(fd, "(");
            dumpVhdlActionExpression(fd, an->left);
            fprintf(fd, " + ");
            dumpVhdlActionExpression(fd, an->right);
            fprintf(fd, ")");
            break;
        case P_AN_MINUS:
            fprintf(fd, "(");
            dumpVhdlActionExpression(fd, an->left);
            fprintf(fd, " - ");
            dumpVhdlActionExpression(fd, an->right);
            fprintf(fd, ")");
            break;
        case P_AN_CONSTANT:
            fprintf(fd, "%d", an->data.constant);
            break;
        case P_AN_SIGNAL:
            fprintf(fd, "INT(%s)", an->data.signal->name);
            break;
        case P_AN_SIGNAL_ARRAY:
            fprintf(fd, "INT(%s(", an->data.signal->name);
            dumpVhdlIndex(fd, an->index);
            fprintf(fd, "))");
            break;
        default:
            fatal_error("dumpVhdlActionExpression()");
    }
}


void dumpVhdlIndex(FILE *fd, indexNodePtr in) {

    switch (in->type) {
        case P_IN_SIGNAL:
            fprintf(fd, "INT(%s)", in->data.signal->name);
            break;
        case P_IN_CONSTANT:
            fprintf(fd, "%d", in->data.constant);
            break;
        default:
            fatal_error("dumpVhdlIndex()");
    }
}


